use crate::error::*;
use crate::http;

use crate::websocket::FrameHeader;
use base64;
use crypto;
use crypto::digest::Digest;

pub fn connection_handler(
    client: &mut http::HttpStream,
    req: &http::Request,
    mut res: http::Response,
) {
    if let http::Method::GET = req.method {
    } else {
        res.status(405);
        res.add_header("Allow", "GET");
        return http::error_handler(
            client,
            res,
            Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("method not allowed".to_string()),
        );
    }

    if let Some(connection) = req.find_header("Connection") {
        if !connection.eq_ignore_ascii_case("upgrade") {
            return http::error_handler(
                client,
                res,
                Error::new(Kind::WebSocketError, Class::ClientError)
                    .set_desc("invalid value for header field 'Connection'".to_string()),
            );
        }
    } else {
        return http::error_handler(
            client,
            res,
            Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("unable to find header field 'Connection'".to_string()),
        );
    }

    if let Some(upgrade) = req.find_header("Upgrade") {
        if !upgrade.eq_ignore_ascii_case("websocket") {
            return http::error_handler(
                client,
                res,
                Error::new(Kind::WebSocketError, Class::ClientError)
                    .set_desc("invalid value for header field 'Upgrade'".to_string()),
            );
        }
    } else {
        return http::error_handler(
            client,
            res,
            Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("unable to find header field 'Upgrade'".to_string()),
        );
    }

    if let Some(version) = req.find_header("Sec-WebSocket-Version") {
        if !version.eq("13") {
            return http::error_handler(
                client,
                res,
                Error::new(Kind::WebSocketError, Class::ClientError)
                    .set_desc("invalid value for header field 'Sec-WebSocket-Key'".to_string()),
            );
        }
    } else {
        return http::error_handler(
            client,
            res,
            Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("unable to find header field 'Sec-WebSocket-Version'".to_string()),
        );
    }

    if let Some(key) = req.find_header("Sec-WebSocket-Key") {
        let mut hasher = crypto::sha1::Sha1::new();
        hasher.input_str(key);
        hasher.input_str("258EAFA5-E914-47DA-95CA-C5AB0DC85B11");

        let mut result = [0u8; 160 / 8];
        hasher.result(&mut result);
        let key = base64::encode(result);
        res.add_header("Sec-WebSocket-Accept", key.as_str());
    } else {
        return http::error_handler(
            client,
            res,
            Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("unable to find header field 'Sec-WebSocket-Key'".to_string()),
        );
    }

    client.server_keep_alive = false;
    res.add_header("Connection", "Upgrade");
    res.add_header("Upgrade", "websocket");

    res.status(101);
    res.send(&mut client.stream).unwrap();

    loop {
        let header = FrameHeader::from(&mut client.stream).unwrap();
        if header.mask {
            println!("Mask: {:X}", header.masking_key.unwrap());
        }

        let mut buf = [0u8; 8192];
        client
            .stream
            .read_exact(&mut buf[..header.payload_len() as usize])
            .unwrap();

        if header.mask {
            let key: [u8; 4] = [
                (header.masking_key.unwrap() >> 24) as u8,
                ((header.masking_key.unwrap() >> 16) & 0xFF) as u8,
                ((header.masking_key.unwrap() >> 8) & 0xFF) as u8,
                (header.masking_key.unwrap() & 0xFF) as u8,
            ];
            for (pos, byte) in buf.iter_mut().enumerate() {
                *byte ^= key[pos % 4];
            }
        }

        println!(
            "Msg: {}",
            String::from_utf8_lossy(&buf[..header.payload_len() as usize])
        );
    }
}
